""" GenerateBarcodeWindow Class """

import cv2
import tkinter
import tkinter.filedialog
from tkinter.messagebox import showerror
import copy
import os
import threading

from kalmus.tkinter_windows.SpecifyMetaDataWindow import SpecifyMetaDataWindow
from kalmus.tkinter_windows.KALMUS_utils import resource_path


class GenerateBarcodeWindow():
    """
    GenerateBarcodeWindow Class
    GUI window for user to generate the barcode from a video file
    """
    def __init__(self, barcode_generator, barcode_stack):
        """
        Initialize
        :param barcode_generator: The barcode generator
        :param barcode_stack: The dictionary that stores all the barcode on the memory
        """
        self.barcode_generator = barcode_generator
        self.barcode_stack = barcode_stack

        # Initialize the window
        self.window = tkinter.Tk()
        self.window.wm_title("Barcode Generator")
        self.window.iconbitmap(resource_path("kalmus_icon.ico"))

        # A temporary meta data dictionary that will hold the meta information given by the user
        self.meta_data_dict = {}

        # video object capture by cv2
        self.video = None

        # Label prompt for the generated barcode's barcode type
        barcode_type_label = tkinter.Label(self.window, text="Barcode Type: ")
        barcode_type_label.grid(row=0, column=0)

        # Label prompt for the frame sampling type
        frame_type_label = tkinter.Label(self.window, text="Frame Type: ")
        frame_type_label.grid(row=1, column=0)

        # Label prompt for the color/brightness metric
        color_metric_label = tkinter.Label(self.window, text="Color metric: ")
        color_metric_label.grid(row=2, column=0)

        # Variable that stores the user's choice of generated barcode type
        self.barcode_type_var = tkinter.StringVar(self.window)
        self.barcode_type_var.set("Color")

        # Dropdown menu for the barcode type selection
        dropdown_bar_type = tkinter.OptionMenu(self.window, self.barcode_type_var, "Color", "Brightness")
        dropdown_bar_type.grid(row=0, column=1)

        # Variable that stores the user's choice of frame sampling type
        self.frame_type_var = tkinter.StringVar(self.window)
        self.frame_type_var.set("Whole_frame")

        # Dropdown menu for the frame sampling type selection
        dropdown_frame_type = tkinter.OptionMenu(self.window, self.frame_type_var, "Whole_frame", "Low_contrast_region",
                                                 "High_contrast_region", "Foreground", "Background")
        dropdown_frame_type.grid(row=1, column=1)

        # Variable that stores the user's choice of color metric
        self.color_metric_var = tkinter.StringVar(self.window)
        self.color_metric_var.set("Average")

        # Dropdown menu for the color metric selection
        dropdown_color_metric = tkinter.OptionMenu(self.window, self.color_metric_var, "Average", "Median", "Mode",
                                                   "Top-dominant", "Weighted-dominant", "Bright", "Brightest")
        dropdown_color_metric.grid(row=2, column=1)

        # Label prompt for the skip over specification
        self.skip_over_label = tkinter.Label(self.window, text="Start at (frames): ")
        self.skip_over_label.grid(row=0, column=2)

        # Label prompt for the sample rate specification
        self.sampled_rate_label = tkinter.Label(self.window, text="Sample every (frames): ")
        self.sampled_rate_label.grid(row=1, column=2)

        # Label prompt for the total frame included specification
        self.total_frames_label = tkinter.Label(self.window, text="Total frames: ")
        self.total_frames_label.grid(row=2, column=2)

        # Acquisition unit specification (in frames or in time)
        self.acquisition_label = tkinter.Label(self.window, text="Acquistion unit")
        self.acquisition_label.grid(row=0, column=5)

        # Variable that stores the user's choice of acquisition unit
        self.acquisition_option = tkinter.StringVar(self.window)
        self.acquisition_option.set("Frame")

        # Radio button for the Acquisition unit (Frame/Time) selection
        radio_frame = tkinter.Radiobutton(self.window, text="Frame", variable=self.acquisition_option,
                                          value="Frame", anchor='w',
                                          command=self.frame_unit)
        radio_frame.grid(row=1, column=5, sticky=tkinter.W)
        radio_frame.select()

        radio_time = tkinter.Radiobutton(self.window, text="Time", variable=self.acquisition_option,
                                         value="Time", anchor='w',
                                         command=self.time_unit)
        radio_time.grid(row=2, column=5, sticky=tkinter.W)

        # Text entry for the skip over specification
        self.skip_over_entry = tkinter.Entry(self.window, textvariable="0", width=12)
        self.skip_over_entry.grid(row=0, column=3, columnspan=2)

        # Text entry for the sampling rate specification
        self.sampled_rate_entry = tkinter.Entry(self.window, textvariable="1", width=12)
        self.sampled_rate_entry.grid(row=1, column=3, columnspan=2)

        # Text entry for the total frames specification
        self.total_frames_entry = tkinter.Entry(self.window, textvariable="1000", width=12)
        self.total_frames_entry.grid(row=2, column=3, columnspan=2)

        # Label prompt for the video file name input
        video_filename_label = tkinter.Label(self.window, text="Media file path: ")
        video_filename_label.grid(row=3, column=0, columnspan=1)

        # Text entry for the video file name specification
        self.video_filename_entry = tkinter.Entry(self.window, textvariable="", width=55)
        self.video_filename_entry.grid(row=3, column=1, columnspan=3)

        # Button to browse the folder
        self.browse_folder_button = tkinter.Button(self.window, text='Browse', command=self.browse_folder)
        self.browse_folder_button.grid(row=3, column=4)

        # Variable that stores the letter box option (automatic/user defined)
        self.letterbox_option = tkinter.StringVar(self.window)
        self.letterbox_option.set("Auto")  # initialize

        # Label prompt for the letter box remove option
        letterbox_label = tkinter.Label(self.window, text="Remove Letterbox: ")
        letterbox_label.grid(row=4, column=0, columnspan=1)

        # Radio button for the letter box remove option
        radio_auto = tkinter.Radiobutton(self.window, text="Auto", variable=self.letterbox_option,
                                         value="Auto", anchor='w',
                                         command=self.disable_setup)
        radio_auto.grid(row=5, column=0, sticky=tkinter.W)
        radio_auto.select()

        radio_manual = tkinter.Radiobutton(self.window, text="Manual", variable=self.letterbox_option,
                                           value="Manual", anchor='w',
                                           command=self.enable_setup)
        radio_manual.grid(row=6, column=0, sticky=tkinter.W)

        # User defined letter box label prompt
        high_ver_label = tkinter.Label(self.window, text="Upper vertical: ")
        high_ver_label.grid(row=5, column=1, columnspan=1)

        low_ver_label = tkinter.Label(self.window, text="Lower vertical: ")
        low_ver_label.grid(row=6, column=1, columnspan=1)

        # User defined letter box text entry
        self.high_ver_entry = tkinter.Entry(self.window, textvariable="-1", width=4, state="disabled")
        self.high_ver_entry.grid(row=5, column=2, columnspan=1, sticky=tkinter.W)

        self.low_ver_entry = tkinter.Entry(self.window, textvariable="-2", width=4, state="disabled")
        self.low_ver_entry.grid(row=6, column=2, columnspan=1, sticky=tkinter.W)

        left_hor_label = tkinter.Label(self.window, text="Left horizontal: ")
        left_hor_label.grid(row=5, column=3, columnspan=1)

        right_hor_label = tkinter.Label(self.window, text="right horizontal: ")
        right_hor_label.grid(row=6, column=3, columnspan=1)

        self.left_hor_entry = tkinter.Entry(self.window, textvariable="-3", width=4, state="disabled")
        self.left_hor_entry.grid(row=5, column=4, columnspan=1)

        self.right_hor_entry = tkinter.Entry(self.window, textvariable="-4", width=4, state="disabled")
        self.right_hor_entry.grid(row=6, column=4, columnspan=1)

        # Variable that stores 0 for not saving frames during the generation 1 for saving frames during the generation
        self.var_saved_frame = tkinter.IntVar(self.window)

        # Checkbox for the user to choose whether save the frames or not during barcode generation
        self.checkbox_saved_frame = tkinter.Checkbutton(self.window, text="Save Frames  ",
                                                        variable=self.var_saved_frame,
                                                        onvalue=1, offvalue=0)
        self.checkbox_saved_frame.grid(row=7, column=0)

        # Variable that stores 0 for single thread generation 1 for multi-thread generation
        self.var_multi_thread = tkinter.IntVar(self.window)

        # Checkbox for the user to choose whether use the multi-thread or not for barcode generation
        self.checkbox_multi_thread = tkinter.Checkbutton(self.window, text="Multi-Thread:",
                                                         variable=self.var_multi_thread,
                                                         onvalue=1, offvalue=0, command=self.update_thread_entry)

        self.checkbox_multi_thread.grid(row=8, column=0)

        # Text entry for the thread specification
        self.thread_entry = tkinter.Entry(self.window, textvariable="-6", width=4, state="normal")
        self.thread_entry.grid(row=8, column=1, sticky=tkinter.W)
        self.thread_entry.delete(0, tkinter.END)
        self.thread_entry.insert(0, "4")
        self.thread_entry.config(state="disabled")

        # Button to generate the barcode
        self.generate_button = tkinter.Button(master=self.window, text="Generate Barcode",
                                              command=self.generate_barcode_thread)
        self.generate_button.grid(row=7, column=2, sticky=tkinter.W, rowspan=2)

        # Button to specify the meta data of the generated barcode
        self.specify_data_button = tkinter.Button(master=self.window, text="Specify Meta Data",
                                                  command=self.specify_data)
        self.specify_data_button.grid(row=7, column=3, sticky=tkinter.W, rowspan=2)

        # Start the window
        self.window.mainloop()

    def specify_data(self):
        """
        Instantiate the SpecifyMetaDataWindow
        :return:
        """
        SpecifyMetaDataWindow(self.meta_data_dict)

    def update_thread_entry(self):
        """
        Enable or disable and change the label prompt when user check/uncheck the multi-thread checkbox
        :return:
        """
        if self.var_multi_thread.get() == 0:
            self.checkbox_multi_thread["text"] = "Multi-Thread:"
            self.thread_entry.config(state="disabled")
        elif self.var_multi_thread.get() == 1:
            self.checkbox_multi_thread["text"] = "# of Threads:"
            self.thread_entry.config(state="normal")

    def time_unit(self):
        """
        Change the acquisition unit to Time when user switch to the time radio button
        :return:
        """
        self.skip_over_label['text'] = "Start at (mins:secs): "
        self.sampled_rate_label['text'] = "Sample every (secs): "
        self.total_frames_label['text'] = "End at (mins:secs): "

    def frame_unit(self):
        """
        Change the acquisition unit to Frame when user switch to the frame radio button
        :return:
        """
        self.skip_over_label['text'] = "Start at (frames): "
        self.sampled_rate_label['text'] = "Sample every (frames): "
        self.total_frames_label['text'] = "Total frames: "

    def browse_folder(self):
        """
        Browse the folder
        :return:
        """
        # Get the file name from the user specification
        filename = tkinter.filedialog.askopenfilename(initialdir=".", title="Select Media file",
                                                      filetypes=(("mp4 files", "*.mp4"), ("avi files", "*.avi"),
                                                                 ("m4v files", "*.m4v"), ("All files", "*.*")))

        # Update the file name to the text entry
        self.video_filename_entry.delete(0, tkinter.END)
        self.video_filename_entry.insert(0, filename)

    def generate_barcode_thread(self):
        threading.Thread(target=self.generate_barcode).start()

    def generate_barcode(self):
        """
        Generate the barcode using the given parameters
        :return:
        """
        # Get barcode type, frame sampling type, and color/brightness metric
        barcode_type = self.barcode_type_var.get()
        frame_type = self.frame_type_var.get()
        color_metric = self.color_metric_var.get()

        # Get the acquisition unit
        unit_type = self.acquisition_option.get()

        # Get the file name/path of the input video
        video_filename = str(self.video_filename_entry.get())

        if not os.path.exists(video_filename):
            showerror("Video File Not Exists", "Video file not found!\nPlease check the file path.")
            return

        # Get the correct acquisition parameters based on the acqusition unit
        if unit_type == "Frame":
            skip_over_str = self.skip_over_entry.get()
            if len(skip_over_str) == 0 or skip_over_str.lower() == "start":
                skip_over = 0
            else:
                skip_over = int(self.skip_over_entry.get())

            total_frames_str = self.total_frames_entry.get()
            if len(total_frames_str) == 0 or total_frames_str.lower() == "end":
                total_frames = int(1e8)
            else:
                total_frames = int(self.total_frames_entry.get())

            sampled_frame_rate = int(self.sampled_rate_entry.get())
        elif unit_type == "Time":
            self.video = cv2.VideoCapture(video_filename)
            fps = self.video.get(cv2.CAP_PROP_FPS)

            skip_over_str = str(self.skip_over_entry.get())
            if len(skip_over_str) == 0 or skip_over_str.lower() == "start":
                skip_over = 0
            else:
                split_pos = skip_over_str.find(":")
                skip_over = int((int(skip_over_str[:split_pos]) * 60 + int(skip_over_str[split_pos + 1:])) * fps)

            sampled_frame_rate_str = str(self.sampled_rate_entry.get())
            sampled_frame_rate = int(round(float(sampled_frame_rate_str) * fps))

            total_frames_str = str(self.total_frames_entry.get())
            if len(total_frames_str) == 0 or total_frames_str.lower() == "end":
                total_frames = int(1e8)
            else:
                split_pos = total_frames_str.find(":")
                total_frames = (int(total_frames_str[:split_pos]) * 60 + int(total_frames_str[split_pos + 1:])) * fps
                total_frames -= skip_over
                total_frames = int(total_frames)
                total_frames //= sampled_frame_rate

        # Make sure the sampled frame rate >= 1 and the skip over >= 0 and total frame >= 0
        if sampled_frame_rate < 1:
            sampled_frame_rate = 1
        if skip_over < 0:
            skip_over = 0
        if total_frames < 0:
            total_frames = 0

        # Update all the parameters to the barcode generator
        self.barcode_generator.barcode_type = barcode_type
        self.barcode_generator.frame_type = frame_type
        self.barcode_generator.color_metric = color_metric
        self.barcode_generator.sampled_frame_rate = sampled_frame_rate
        self.barcode_generator.skip_over = skip_over
        self.barcode_generator.total_frames = total_frames

        # Update the generate button text to the processing
        self.generate_button["text"] = "  Processing... "

        # Check if user choose the multi-thread or not
        if self.var_multi_thread.get() == 0:
            multi_thread = None
        elif self.var_multi_thread.get() == 1:
            # If user choose to use the multi-thread, then get the number of threads that will be used
            multi_thread = int(self.thread_entry.get())

        # Check if user choose to save the frames or not
        if self.var_saved_frame.get() == 1:
            save_frames = True
        else:
            save_frames = False

        # Check if user choose to define the letter box region manually
        if self.letterbox_option.get() == "Manual":
            # Update the letter box parameters, if user choose Manual
            high_ver = int(self.high_ver_entry.get())
            low_ver = int(self.low_ver_entry.get())
            left_hor = int(self.left_hor_entry.get())
            right_hor = int(self.right_hor_entry.get())
            # Start the generation
            self.barcode_generator.generate_barcode(video_filename, user_defined_letterbox=True,
                                                    low_ver=low_ver, high_ver=high_ver,
                                                    left_hor=left_hor, right_hor=right_hor,
                                                    num_thread=multi_thread, save_frames=save_frames)
        elif self.letterbox_option.get() == "Auto":
            # If not, start the generation. The letter box will be automatically found during the generation process
            self.barcode_generator.generate_barcode(video_filename, num_thread=multi_thread, save_frames=save_frames)

        # Correct the total frames
        total_frames = self.barcode_generator.get_barcode().total_frames

        # Get the key of the barcode, which will be later stored in the memory stack (dictionary)
        start_pos = video_filename.rfind("/") + 1
        if start_pos < 0:
            start_pos = 0
        end_pos = video_filename.rfind(".")
        videoname = video_filename[start_pos:end_pos] + "_" + barcode_type + "_" + frame_type + "_" + color_metric \
                    + "_" + str(skip_over) + "_" + str(sampled_frame_rate) + "_" + str(total_frames)

        # Get the barcode from the barcode generator
        barcode = self.barcode_generator.get_barcode()

        # Clear the cv2 captured video object
        barcode.video = None

        # Update the user pre-defined meta data to the computed barcode
        barcode.meta_data = copy.deepcopy(self.meta_data_dict)

        # Add the generated barcode to the memory stack (dictionary)
        self.barcode_stack[videoname] = copy.deepcopy(barcode)

        # Quit the window
        self.window.destroy()

    def disable_setup(self):
        """
        Disable the letter box setup entry if user choose the Auto radio button
        :return:
        """
        self.high_ver_entry.config(state="disabled")
        self.low_ver_entry.config(state="disabled")
        self.left_hor_entry.config(state="disabled")
        self.right_hor_entry.config(state="disabled")

    def enable_setup(self):
        """
        Enable the letter box setup entry if user choose the Manual radio button
        :return:
        """
        self.high_ver_entry.config(state="normal")
        self.low_ver_entry.config(state="normal")
        self.left_hor_entry.config(state="normal")
        self.right_hor_entry.config(state="normal")
